package data;

import java.io.Serializable;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import org.apache.commons.lang3.StringUtils;

import data.Enums.Msd;
import gui.ValidationUtil;

public class Word implements Serializable {
	public static final char PAD_CHARACTER = '-';

	private String word;
	private String lemma;
	private String msd;
	private String normalizedWord;
	private final HashSet<Character> VOWELS = new HashSet<>(Arrays.asList('a', 'e', 'i', 'o', 'u'));

	/**
	 * Possible values:
	 * <p>
	 * <ul>
	 * <li>S = samostalnik</li>
	 * <li>G = glagol</li>
	 * <li>P = pridevnik</li>
	 * <li>R = prislov</li>
	 * <li>Z = zaimek</li>
	 * <li>K = števnik</li>
	 * <li>D = predlog</li>
	 * <li>V = veznik</li>
	 * <li>L = členek</li>
	 * <li>M = medmet</li>
	 * <li>O = okrajšava</li>
	 * <li>N = neuvrščeno</li>
	 * </ul>
	 */
	//private char besedna_vrsta;
	public Word(String word, String lemma, String msd) {
		this.lemma = lemma;
		this.msd = msd; //normalizeMsd(msd);
		this.normalizedWord = "";

		// veliko zacetnico ohranimo samo za lastna imena
		if (!ValidationUtil.isEmpty(this.msd) && !(this.msd.charAt(0) == 'S'
				&& this.msd.length() >= 2
				&& this.msd.charAt(1) == 'l')) {
			this.word = word.toLowerCase();
		} else {
			this.word = word;
		}
	}

	public Word(String word, String lemma, String msd, String normalizedWord) {
		this.lemma = lemma;
//		this.msd = normalizeMsd(msd);
		this.msd = msd;
		this.normalizedWord = normalizedWord;

		// veliko zacetnico ohranimo samo za lastna imena
		if (!ValidationUtil.isEmpty(this.msd) && !(this.msd.charAt(0) == 'S'
				&& this.msd.length() >= 2
				&& this.msd.charAt(1) == 'l')) {
			this.word = word.toLowerCase();
		} else {
			this.word = word;
		}
	}

	public Word() {
	}

//	/**
//	 * Appends a number of '-' to msds which are not properly sized.
//	 * E.g. nouns should have 5 attributes, yet the last one isn't always defined (Somei vs. Sometd)
//	 *
//	 * @param msdInput
//	 *
//	 * @return
//	 */
//	private String normalizeMsd(String msdInput) {
//		if (ValidationUtil.isEmpty(msdInput)) {
//			return "";
//		} else {
//			return StringUtils.rightPad(msdInput, Msd.getMsdLengthForType(msdInput), PAD_CHARACTER);
//		}
//	}

	public Word(String word) {
		this.word = word;
	}

	public String getWord() {
		return word;
	}

	public String getCVVWord() {
		return covertToCvv(word);
	}

	public String getCVVLemma() {
		return covertToCvv(lemma);
	}

	private String covertToCvv(String s) {
		char[] StringCA = s.toCharArray();

		for (int i = 0; i < StringCA.length; i++) {
			StringCA[i] = VOWELS.contains(StringCA[i]) ? 'V' : 'C';
		}

		return new String(StringCA);
	}

	public void setWord(String word) {
		this.word = word;
	}

	public String getLemma() {
		return lemma;
	}

	public void setLemma(String lemma) {
		this.lemma = lemma;
	}

	public String getMsd() {
		return msd;
	}

	public void setMsd(String msd) {
		this.msd = msd;
	}

	public String getNormalizedWord() {
		return normalizedWord;
	}

	public void setNormalizedWord(String normalizedWord) {
		this.normalizedWord = normalizedWord;
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();

		sb.append("beseda:\t")
				.append(getWord())
				.append("\n")
				.append("lema:\t")
				.append(getLemma())
				.append("\n")
				.append("msd:\t")
				.append(getMsd())
				.append("normalized word:\t")
				.append(getNormalizedWord())
				.append("\n");

		return sb.toString();
	}

	public String getForCf(CalculateFor calculateFor, boolean cvv) {
		String returnValue = "";

		if (cvv) {
			returnValue = calculateFor == CalculateFor.WORD ? getCVVWord() : getCVVLemma();
		} else {
			returnValue = calculateFor == CalculateFor.WORD ? getWord() : getLemma();
		}

		return returnValue;
	}
}
